@@ -22,20 +22,20 @@
  */
 
 #include "rcs.h"
-RCS_ID("$Id: termcap.c,v 1.8 1994/05/31 12:33:13 mlschroe Exp $ FAU")
+RCS_ID("$Id: termcap.c,v 1.9 1994/09/06 17:00:31 mlschroe Exp $ FAU")
 
 #include <sys/types.h>
 #include "config.h"
 #include "screen.h"
 #include "extern.h"
 
-
 extern struct display *display, *displays;
 extern int real_uid, real_gid, eff_uid, eff_gid;
 extern struct term term[];	/* terminal capabilities */
 extern struct NewWindow nwin_undef, nwin_default, nwin_options;
 extern int force_vt, assume_LP;
 extern int Z0width, Z1width;
+extern int hardstatusemu;
 #ifdef MAPKEYS
 extern struct action umtab[];
 extern struct action mmtab[];
@@ -45,6 +45,7 @@ extern int kmap_extras_fl[];
 extern int DefaultEsc;
 #endif
 
+
 static void  AddCap __P((char *));
 static void  MakeString __P((char *, char *, int, char *));
 static char *findcap __P((char *, char **, int));
@@ -56,7 +57,7 @@ static int   e_tgetnum __P((char *));
 #ifdef MAPKEYS
 static int   addmapseq __P((char *, int));
 static int   remmapseq __P((char *));
-#ifdef DEBUG
+#ifdef DEBUGG
 static void  dumpmap __P((void));
 #endif
 #endif
@@ -94,6 +95,11 @@ char *s;
   return 0;
 }
 
+/*
+ * Compile the terminal capabilities for a display.
+ * Input: tgetent(, D_termname) extra_incap, extra_outcap.
+ * Effect: display initialisation.
+ */
 int
 InitTermcap(wi, he)
 int wi;
@@ -116,7 +122,7 @@ int he;
 #endif
       return -1;
     }
-  debug1("got it:\n%s\n",tbuf);
+  debug1("got it:\n%s\n", tbuf);
 #ifdef DEBUG
   if (extra_incap)
     debug1("Extra incap: %s\n", extra_incap);
@@ -130,6 +136,9 @@ int he;
       return -1;
     }
 
+  /*
+   * loop through all needed capabilities, record their values in the display
+   */
   tp = D_tentry;
   for (i = 0; i < T_N; i++)
     {
@@ -152,6 +161,10 @@ int he;
 	  /*NOTREACHED*/
 	}
     }
+
+  /*
+   * Now a good deal of sanity checks on the retrieved capabilities.
+   */
   if (D_HC)
     {
       Msg(0, "You can't run screen on a hardcopy terminal.");
@@ -189,8 +202,8 @@ int he;
     {
       /* standard fixes for xterms etc */
       /* assume color if name doesn't begin with vt but it
-	 looks like a vt */
-      if (!D_CAF && D_ME && InStr(D_ME, "\033[m"))
+         looks like a vt */
+      if (!D_CAF && strncmp(D_termname, "vt", 2) && D_ME && InStr(D_ME, "\033[m"))
 	{
 #ifdef TERMINFO
 	  D_CAF = "\033[3%p1%dm";
@@ -201,14 +214,13 @@ int he;
 #endif
 	}
       if (D_OP && InStr(D_OP, "\033[39;49m"))
-	D_CAX = 1;
+        D_CAX = 1;
       if (strcmp(D_termname, "xterm-color") == 0)
 	D_BE = 1;
       /* ISO2022 */
       if ((D_EA && InStr(D_EA, "\033(B")) || (D_AS && InStr(D_AS, "\033(0")))
 	D_CG0 = 1;
     }
-
   if (nwin_options.flowflag == nwin_undef.flowflag)
     nwin_default.flowflag = D_CNF ? FLOW_NOW * 0 : 
 			    D_NX ? FLOW_NOW * 1 :
@@ -315,8 +327,8 @@ int he;
       D_CAF = D_CSF;
       D_CAB = D_CSB;
     }
-  if (D_BE)	/* screen erased with background color */
-    D_UT = 1;
+  if (D_BE)
+    D_UT = 1;	/* screen erased with background color */
 
   if (!D_DO)
     D_DO = D_NL;
@@ -337,6 +349,8 @@ int he;
     D_VI = D_VS = 0;
   if (D_CCE == 0)
     D_CCS = 0;
+
+#ifdef FONT
   if (D_CG0)
     {
       if (D_CS0 == 0)
@@ -376,13 +390,16 @@ int he;
     for (i = strlen(D_CC0) & ~1; i >= 0; i -= 2)
       D_c0_tab[(int)(unsigned char)D_CC0[i]] = D_CC0[i + 1];
   debug1("ISO2022 = %d\n", D_CG0);
+#endif /* FONT */
   if (D_PF == 0)
     D_PO = 0;
   debug2("terminal size is %d, %d (says TERMCAP)\n", D_CO, D_LI);
 
+#ifdef FONT
   if (D_CXC)
     if (CreateTransTable(D_CXC))
       return -1;
+#endif
 
   /* Termcap fields Z0 & Z1 contain width-changing sequences. */
   if (D_CZ1 == 0)
@@ -400,6 +417,10 @@ int he;
       if (D_WS <= 0)
         D_WS = D_width;
     }
+  D_has_hstatus = hardstatusemu & ~HSTATUS_ALWAYS;
+  if (D_HS && !(hardstatusemu & HSTATUS_ALWAYS))
+    D_has_hstatus = HSTATUS_HS;
+
 #ifdef KANJI
   D_kanji = 0;
   if (D_CKJ)
@@ -411,7 +432,7 @@ int he;
     }
 #endif
   if (!D_tcs[T_NAVIGATE].str && D_tcs[T_NAVIGATE + 1].str)
-    D_tcs[T_NAVIGATE].str = D_tcs[T_NAVIGATE + 1].str;	/* kh = @1 */
+    D_tcs[T_NAVIGATE].str = D_tcs[T_NAVIGATE + 1].str;  /* kh = @1 */
   if (!D_tcs[T_NAVIGATE + 2].str && D_tcs[T_NAVIGATE + 3].str)
     D_tcs[T_NAVIGATE + 2].str = D_tcs[T_NAVIGATE + 3].str; /* kH = @7 */
 
@@ -435,6 +456,7 @@ int he;
     {
       debug1("termcap has OL (%d), setting limit\n", D_COL);
       D_obufmax = D_COL;
+      D_obuflenmax = D_obuflen - D_obufmax;
     }
 #ifdef MAPKEYS
   D_nseqs = 0;
@@ -525,13 +547,13 @@ CheckEscape()
   for (display = displays; display; display = display->d_next)
     {
       for (i = 0; i < D_nseqs; i++)
-	{
-	  nr = D_kmaps[i].nr & ~KMAP_NOTIMEOUT;
-	  if (umtab[nr].nr == RC_COMMAND)
-	    break;
-	  if (umtab[nr].nr == RC_ILLEGAL && dmtab[nr].nr == RC_COMMAND)
-	    break;
-	}
+        {
+          nr = D_kmaps[i].nr & ~KMAP_NOTIMEOUT;
+          if (umtab[nr].nr == RC_COMMAND)
+            break;
+          if (umtab[nr].nr == RC_ILLEGAL && dmtab[nr].nr == RC_COMMAND)
+            break;
+        }
       if (i >= D_nseqs)
         break;
     }
@@ -541,8 +563,10 @@ CheckEscape()
       return;
     }
   ParseEscape((struct user *)0, "^aa");
-  if (odisplay->d_user->u_Esc <= 0)
+  if (odisplay->d_user->u_Esc == -1)
     odisplay->d_user->u_Esc = DefaultEsc;
+  if (odisplay->d_user->u_MetaEsc == -1)
+    odisplay->d_user->u_MetaEsc = DefaultMetaEsc;
   display = 0;
   Msg(0, "Warning: escape char set back to ^A");
   display = odisplay;
@@ -602,7 +626,7 @@ int nr;
 	  continue;
 	(*o)++;
       }
-#ifdef DEBUG
+#ifdef DEBUGG
   dumpmap();
 #endif
   return 0;
@@ -631,13 +655,13 @@ char *seq;
   for (j = i + 1; j < D_nseqs; j++)
     D_kmaps[j - 1] = D_kmaps[j];
   D_nseqs--;
-#ifdef DEBUG
+#ifdef DEBUGG
   dumpmap();
 #endif
   return 0;
 }
 
-#ifdef DEBUG
+#ifdef DEBUGG
 static void
 dumpmap()
 {
@@ -659,10 +683,13 @@ dumpmap()
       debug2(" ==> %d%s\n", n & ~KMAP_NOTIMEOUT, (n & KMAP_NOTIMEOUT) ? " (no timeout)" : "");
     }
 }
-#endif
+#endif /* DEBUGG */
 
-#endif
+#endif /* MAPKEYS */
 
+/*
+ * Appends to the static variable Termcap
+ */
 static void
 AddCap(s)
 char *s;
@@ -685,6 +712,10 @@ char *s;
     Panic(0, "TERMCAP overflow - sorry.");
 }
 
+/*
+ * Reads a displays capabilities and reconstructs a termcap entry in the 
+ * global buffer "Termcap". A pointer to this buffer is returned.
+ */
 char *
 MakeTermcap(aflag)
 int aflag;
@@ -750,13 +781,13 @@ int aflag;
     if (tgetstr("im", &xbp) && tgetstr("ic", &xbp) && displays)
       {
 #ifdef TERMINFO
-	Msg(0, "Warning: smir and ich1 set in %s terminfo entry", p);
+        Msg(0, "Warning: smir and ich1 set in %s terminfo entry", p);
 #else
-	Msg(0, "Warning: im and ic set in %s termcap entry", p);
+        Msg(0, "Warning: im and ic set in %s termcap entry", p);
 #endif
       }
   }
-  
+
   tcLineLen = 100;	/* Force NL */
   if (strlen(Term) > TERMCAP_BUFSIZE - 40)
     strcpy(Term, "too_long");
@@ -862,7 +893,7 @@ int aflag;
 	}
 #endif
       if (D_CG0)
-        AddCap("G0:");
+	AddCap("G0:");
       if (D_CC0 || (D_CS0 && *D_CS0))
 	{
 	  AddCap("as=\\E(0:");
@@ -993,6 +1024,7 @@ char *s;
 #define QUOTES(p) \
   (*p == '\\' && (p[1] == '\\' || p[1] == ',' || p[1] == '%'))
 
+#ifdef FONT
 int
 CreateTransTable(s)
 char *s;
@@ -1079,7 +1111,7 @@ char *s;
 	      *sx++ = *p;
 	    }
 	  *sx = 0;
-	  ASSERT(ctable[c] + l == sx);
+	  ASSERT(ctable[c] + l * templnsub + templlen == sx);
 	  debug3("XC: %c %c->%s\n", curchar, c, ctable[c]);
 	}
       if (*s == ',')
@@ -1104,10 +1136,11 @@ FreeTransTable()
       for (j = 0; j < 257; j++, q++)
 	if (*q)
 	  free(*q);
-      free((char *)*p);
+      free(*p);
     }
-  free((char *)D_xtable);
+  free(D_xtable);
 }
+#endif /* FONT */
 
 static int
 copyarg(pp, s)
